
__version_info__ = ('0', '1', '38')
__version__ = '.'.join(__version_info__)
__authors__ = "Silvio Giancola"
__authors_username__ = "giancos"
__author_email__ = "silvio.giancola@kaust.edu.sa"
__github__ = 'https://github.com/SilvioGiancola/SoccerNetv2'


# import logging
# from SoccerNet.Downloader import SoccerNetDownloader
# from SoccerNet.utils import getListGames

# from SoccerNet.Evaluation import ActionSpotting

# try:
#     from SoccerNet.DataLoader import VideoLoader
# except:
#     logging.info("Install TensorFlow/PyTorch for VideoLoader")
#     pass


# try:
#     from SoccerNet.Feature import Extractor
# except:
#     # print("Install TensorFlow/PyTorch for FeatureExtractor")
#     # logging.debug("Install TensorFlow/PyTorch for FeatureExtractor")
#     # print("Install TensorFlow/PyTorch for FeatureExtractor")
#     pass


# # try:
# # except:
# #     logging.info("Install TensorFlow for FeatureExtractor")
# #     pass
    
# # try:
# #     from SoccerNet.DataLoaderTorch import SoccerNetDataLoaderTorch
# # except:
# #     logging.info("Install Torch for SoccerNetDataLoaderTorch")
# #     pass


# try:
#     from SoccerNet.DataLoaderTensorFlow import SoccerNetDataLoaderTensorFlow
# except:
#     logging.info("Install TensorFlow for SoccerNetDataLoaderTensorFlow")
#     pass


